VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "cScrollbar"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

Private Declare Function GetCursorPos Lib "user32" (lpPoint As POINTAPI) As Long
Private Declare Function GetCapture Lib "user32" () As Long
Private Declare Function SetCapture Lib "user32" (ByVal hWnd As Long) As Long
Private Declare Function WindowFromPoint Lib "user32" (ByVal xpoint As Long, ByVal ypoint As Long) As Long

Private Type POINTAPI
    X  As Long
    Y  As Long
End Type

Private Enum enumElement
   eNone
   eTopButton
   eBottomButton
   eScrollThumb
   eScrollTrackAbove
   eScrollTrackBelow
End Enum

Private WithEvents mHost As PictureBox, WithEvents mMouseDownTimer As cTimer, WithEvents mMouseLeaveTimer As cTimer
Attribute mHost.VB_VarHelpID = -1
Attribute mMouseDownTimer.VB_VarHelpID = -1
Attribute mMouseLeaveTimer.VB_VarHelpID = -1
Private mCC As cCairoContext, mColour As Long, mHotColour As Long, mButtonColour As Long
Private mMin As Double, mMax As Double, mSmallChange As Double, mLargeChange As Double, mValue As Double, mPreDragValue As Double
Private mScrollThumbY As Double, mScrollThumbH As Double
Private mMouseX As Single, mMouseY As Single, mScrollTrackMouseDownY As Double, mMouseDownElement As enumElement, mMouseOverElement As enumElement
Private mVertical As Boolean, mWidth As Double, mHeight As Double, mYScale As Double

Event Change(Value As Double) 'both events are only raised via user interation...
Event Scroll(ExactValue As Double) '...setting the .Value property in code does not cause a change event! This is intentional.
Private Sub Class_Initialize()
   Set mMouseDownTimer = New_c.Timer()
   Set mMouseLeaveTimer = New_c.Timer(50)
   mColour = vbWindowBackground: mButtonColour = mColour: mHotColour = vbHighlight
End Sub
Private Sub Class_Terminate()
   Set mCC = Nothing
   Set mHost = Nothing
   Set mMouseDownTimer = Nothing
   Set mMouseLeaveTimer = Nothing
End Sub
Public Sub Init(pHost As PictureBox, pVertical As Boolean, pMin As Double, pMax As Double, pSmallChange As Double, pLargeChange As Double)
   If pVertical Then pHost.Width = Cairo.Theme.GetScrollerSize Else pHost.Height = Cairo.Theme.GetScrollerSize
   mVertical = pVertical
   Set mHost = pHost
   SetParams pMin, pMin, pMax, pSmallChange, pLargeChange
End Sub
Public Sub SetParams(pValue As Double, pMin As Double, pMax As Double, pSmallChange As Double, pLargeChange As Double)
   mMin = pMin: mMax = IIf(pMax < pMin, pMin, pMax)
   mSmallChange = IIf(pSmallChange > mMax - mMin, 1, pSmallChange)
   mLargeChange = pLargeChange
   mValue = IIf(pValue < mMin Or pValue > mMax, mMin, pValue)
   mHost_Resize
End Sub
Public Sub SetColours(pScrollTrack As Long, pButtons As Long, pHotColour As Long)
   mColour = pScrollTrack: mButtonColour = pButtons: mHotColour = pHotColour
   Draw
End Sub
Public Property Let Value(pValue As Double)
   If Not mMouseDownElement = eScrollThumb Then UpdateValue pValue 'don't allow the internal value to be changed from outside while thumb is being dragged
End Property
Public Property Get Value() As Double
   Value = mValue
End Property
Public Property Let Min(pMin As Double)
   SetParams mValue, pMin, mMax, mSmallChange, mLargeChange
End Property
Public Property Let Max(pMax As Double)
   SetParams mValue, mMin, pMax, mSmallChange, mLargeChange
End Property
Public Property Let SmallChange(pSmallChange As Double)
   SetParams mValue, mMin, mMax, pSmallChange, mLargeChange
End Property
Public Property Let LargeChange(pLargeChange As Double)
   SetParams mValue, mMin, mMax, mSmallChange, pLargeChange
End Property
Public Property Get Max() As Double: Max = mMax: End Property
Public Property Get Min() As Double: Min = mMin: End Property
Public Property Get SmallChange() As Double: SmallChange = mSmallChange: End Property
Public Property Get LargeChange() As Double: LargeChange = mLargeChange: End Property
Private Sub UpdateValue(pNewValue As Double, Optional NotifyChange As Boolean, Optional ForceNotify As Boolean)
   If pNewValue < mMin Then pNewValue = mMin
   If pNewValue > mMax Then pNewValue = mMax
   If mValue = pNewValue And Not ForceNotify Then Exit Sub
   mValue = pNewValue
   Draw
   If NotifyChange Then RaiseEvent Change((mValue))
End Sub
Private Sub Draw()
   mScrollThumbY = mWidth + (mValue - mMin) * mYScale
   DrawScrollTrack
   DrawArrowButton eTopButton, 0
   DrawArrowButton eBottomButton, mHeight - mWidth - 1
   If Not mMin = mMax Then DrawScrollThumb
   Set mHost.Picture = mCC.Surface.Picture
End Sub
Private Sub DrawScrollTrack()
   mCC.SetSourceColor mColour: mCC.Paint
   If mMouseDownElement > 3 Then
      mCC.Rectangle 0, mScrollThumbY + mScrollThumbH / 2, mWidth, IIf(mMouseDownElement = eScrollTrackAbove, -mHeight, mHeight)
      mCC.SetSourceColor mColour: mCC.Fill 'True
   End If
End Sub
Private Sub DrawArrowButton(Element As enumElement, Y As Double)
   With mCC
      .Rectangle 0, Y, mWidth, mWidth  ', True
      .SetSourceColor mColour: .Fill mMouseDownElement = Element Or mMouseOverElement = Element Or mMouseDownElement = Element + 3
      If mMouseDownElement = Element Then
        .SetSourceColor mHotColour, 0.4: .Fill
      ElseIf mMouseOverElement = Element Or mMouseDownElement = Element + 3 Then
        .SetSourceColor mHotColour, 0.2: .Fill
      End If

      .Save
         If Element = eBottomButton Then .TranslateDrawings 0, mHeight - mWidth
         
         If Element = eBottomButton Then
            .MoveTo mWidth * 0.2, mWidth * 0.33
            .LineTo mWidth * 0.5, mWidth * 0.66
            .LineTo mWidth * 0.8, mWidth * 0.33
         Else
            .MoveTo mWidth * 0.2, mWidth * 0.66
            .LineTo mWidth * 0.5, mWidth * 0.33
            .LineTo mWidth * 0.8, mWidth * 0.66
         End If
         .SetSourceColor vbBlack
         .SetLineWidth mWidth / 10
         .Stroke
      .Restore
   End With
End Sub
Private Sub DrawScrollThumb()
  mCC.Rectangle 0, mScrollThumbY, mWidth, mScrollThumbH  ', True
  mCC.SetSourceColor mColour: mCC.Fill True
  If mMouseDownElement = eScrollThumb Then
    mCC.SetSourceColor mHotColour, 0.4: mCC.Fill
  ElseIf mMouseOverElement = eNone And mMouseDownElement = eNone Then
    mCC.SetSourceColor mHotColour, 0.2: mCC.Fill
  ElseIf mMouseOverElement > eNone Or mMouseDownElement > eNone Then
    mCC.SetSourceColor mHotColour, 0.3: mCC.Fill
  Else
    mCC.Stroke
  End If
End Sub
Private Sub mHost_DblClick() 'rapid, successive mouse-downs cause double-clicks (rather than 2 mouse-downs) so we handle that here
   SetCapture mHost.hWnd
   If mVertical Then
      mHost_MouseDown vbLeftButton, 0, mMouseX, mMouseY
   Else 'need to 'undo' our stored values before passing them into a new mouse-related function (as they will change there)
      mHost_MouseDown vbLeftButton, 0, mMouseY, -mMouseX + mWidth
   End If
End Sub
Private Sub mHost_MouseDown(Button As Integer, Shift As Integer, X As Single, Y As Single)
   If Not Button = vbLeftButton Then Exit Sub
   If mVertical Then mMouseX = X: mMouseY = Y Else mMouseX = -Y + mWidth: mMouseY = X
   mMouseDownElement = HitTest(mMouseX, mMouseY)
   Draw
   If mMouseDownElement = eScrollThumb Then
      mScrollTrackMouseDownY = mMouseY - mScrollThumbY
      mPreDragValue = mValue
   End If
   mMouseDownTimer_Timer
End Sub
Private Sub mHost_MouseMove(Button As Integer, Shift As Integer, X As Single, Y As Single)
Dim PrevMouseOverElement As enumElement
   mMouseLeaveTimer.Enabled = mMouseDownElement = eNone
   If mVertical Then mMouseX = X: mMouseY = Y Else mMouseX = -Y + mWidth: mMouseY = X
   If mMouseDownElement = eNone Then mMouseOverElement = HitTest(mMouseX, mMouseY)
   If mMouseOverElement <> PrevMouseOverElement Then PrevMouseOverElement = mMouseOverElement: Draw
   If Button = vbLeftButton And mMouseDownElement = eScrollThumb Then
      If mMouseX > -140 And mMouseX < mWidth + 140 Then
         UpdateValue mValue + ((mMouseY - mScrollTrackMouseDownY) / mYScale) - (mScrollThumbY / mYScale)
      Else
         UpdateValue mPreDragValue
      End If
      RaiseEvent Scroll(mValue)
   Else
      If Not Button = vbLeftButton Then mMouseDownElement = eNone 'belt and braces!
   End If
End Sub
Private Sub mHost_MouseUp(Button As Integer, Shift As Integer, X As Single, Y As Single)
Dim ForceChangeEvent As Boolean
   ForceChangeEvent = mMouseDownElement = eScrollThumb And mPreDragValue <> Round(mValue)
   mMouseDownElement = eNone
   Draw
   mMouseDownTimer.Enabled = False: mMouseLeaveTimer.Enabled = True
   UpdateValue Round(mValue), ForceChangeEvent, ForceChangeEvent
End Sub
Private Sub mHost_Resize()
   Set mCC = Cairo.CreateSurface(mHost.ScaleWidth, mHost.ScaleHeight).CreateContext
   If mVertical Then
      mWidth = mHost.ScaleWidth: mHeight = mHost.ScaleHeight
   Else
      mWidth = mHost.ScaleHeight: mHeight = mHost.ScaleWidth
      mCC.TranslateDrawings 0, mWidth
      mCC.RotateDrawingsDeg -90
   End If
   
   If mMin <> mMax Then
      mScrollThumbH = ((mHeight - mWidth * 2) / (mMax - mMin + mLargeChange)) * mLargeChange
      If mScrollThumbH < 16 Then mScrollThumbH = 16
      mYScale = (mHeight - mScrollThumbH - mWidth * 2) / (mMax - mMin)
      mHost.Enabled = True
   Else
      mScrollThumbH = (mHeight - mWidth * 2)
      mHost.Enabled = False
   End If
   Draw
End Sub
Private Function HitTest(X As Single, Y As Single) As enumElement
   Select Case True
      Case Y <= mWidth And X > 0 And X < mWidth
         HitTest = eTopButton
      Case Y >= mHeight - mWidth And X > 0 And X < mWidth
         HitTest = eBottomButton
      Case Y >= mScrollThumbY And Y <= mScrollThumbY + mScrollThumbH
         HitTest = eScrollThumb
      Case Y < mScrollThumbY And X > 0 And X < mWidth
         HitTest = eScrollTrackAbove
      Case Y > mScrollThumbY + mScrollThumbH And X > 0 And X < mWidth
         HitTest = eScrollTrackBelow
      Case Else
         HitTest = eNone
   End Select
End Function
Private Sub mMouseLeaveTimer_Timer()
Dim pt As POINTAPI, MouseLeft As Boolean
   GetCursorPos pt
   MouseLeft = WindowFromPoint(pt.X, pt.Y) <> mHost.hWnd And mMouseDownElement = eNone
   mMouseLeaveTimer.Enabled = Not MouseLeft
   If MouseLeft And mMouseDownElement = eNone Then mMouseOverElement = eNone: Draw
End Sub
Private Sub mMouseDownTimer_Timer()
Dim MouseLostCapture As Boolean
   mMouseOverElement = HitTest(mMouseX, mMouseY) 'need to check this as it can change even when the mouse is still (i.e. the thumb moving)
   If Not mMouseOverElement = mMouseDownElement Then
      Draw
   Else
      Select Case mMouseDownElement
         Case eTopButton: UpdateValue mValue - mSmallChange, True
         Case eBottomButton: UpdateValue mValue + mSmallChange, True
         
         Case eScrollTrackAbove: UpdateValue mValue - mLargeChange, True
         Case eScrollTrackBelow: UpdateValue mValue + mLargeChange, True
      End Select
   End If
   MouseLostCapture = GetCapture <> mHost.hWnd
   If MouseLostCapture Then mMouseDownElement = eNone:  Draw
   mMouseDownTimer.Interval = IIf(mMouseDownTimer.Enabled, 50, 500) 'handle a repeat delay
   mMouseDownTimer.Enabled = Not MouseLostCapture: mMouseLeaveTimer.Enabled = MouseLostCapture
End Sub
