VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "cExposedNodes"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit
Private WithEvents mAllNodes                As cVhNodes 'reference to the entire disorganised node collection
Attribute mAllNodes.VB_VarHelpID = -1
Private mExposedNodes                       As Collection 'holds an ordered sub-set of the above
Private mNotifyFrequency                    As Double
Private mNodeIndentation                    As Long
Private mWidestNodeWidth                    As Long
Private mWidestNodeKey                      As String
Private mRecalcWidestNode                   As Boolean

'Events raised to clsTreeViewport
Event CompletelyRebuilt()
Event NodesRemoved()
Event NodesAdded(ParentNodeKey As String, ExposedDescendentCount As Long)
Private Sub mAllNodes_NodeCaptionChanged(NodeKey As String)
    RaiseEvent CompletelyRebuilt
End Sub
Private Sub mAllNodes_NodesCleared()
    Rebuild
End Sub
Private Sub mAllNodes_NodesSorted()
   Rebuild
End Sub
Private Sub mAllNodes_NodeMoved()
   Rebuild
End Sub
Private Sub mAllNodes_NodeRemoved(RemovedNode As cVhNode)
'    Debug.Print "Removing " & RemovedNode.Caption & "(" & RemovedNode.NodeType & ")"
    
    If RemovedNode.IsExposed Then
        If RemovedNode.Expanded Then
            CascadeRemove RemovedNode.Key, RemovedNode.Children, True, 0
        End If
        Remove RemovedNode
    End If
    
    If mRecalcWidestNode Then CalcMaxNodeWidth
    
    RaiseEvent CompletelyRebuilt
End Sub
Private Sub mAllNodes_NodeAdded(NewNode As cVhNode)

    'if the new node is exposed (all ancestors expanded), then add it to the mExposedNodes collection too
    If NewNode.Parent.IsExposed Then
        If NewNode.Parent.Expanded And NewNode.Parent.NodeType.SortMethod = eNoSort Then
            Add NewNode
            RaiseEvent CompletelyRebuilt
        Else
            If NewNode.Parent.Children.Count = 1 Then
                RaiseEvent CompletelyRebuilt
            End If
        End If
    End If

End Sub
Private Sub mAllNodes_NodeExpanded(pNodeKey As String, pCascade As Boolean)
Dim thisNode As cVhNode, lNumNodesRevealed As Long, lVisibleDescendentCount As Long, CurrentlyExpanded As Boolean
    
    If pNodeKey = MASTER_ROOT Then
        Set thisNode = mAllNodes.RootNode
    Else
        Set thisNode = mAllNodes(pNodeKey)
    End If
    
    If (Not thisNode.Expanded And thisNode.Children.Count > 0) Or pCascade Then
    
        Debug.Print vbCrLf & "EXPANDING "; thisNode.Caption
        
        CurrentlyExpanded = thisNode.Expanded
        thisNode.Expanded = True
        
        CascadeAdd pNodeKey, (pNodeKey), thisNode.Children, CurrentlyExpanded, pCascade, lNumNodesRevealed, lVisibleDescendentCount
        
        If lNumNodesRevealed > 0 Then
            RaiseEvent NodesAdded(pNodeKey, lVisibleDescendentCount)
        End If
    
        Debug.Print "EXPANSION COMPLETE: " & lNumNodesRevealed & " nodes revealed"
    End If

End Sub
Private Sub mAllNodes_NodeCollapsed(pNodeKey As String, pCascade As Boolean)
Dim thisNode As cVhNode, lNodesConcealedCount As Long
    
    lNodesConcealedCount = 0
    
    If pNodeKey = MASTER_ROOT Then
        Set thisNode = mAllNodes.RootNode
        mAllNodes.SelectNode 1
    Else
        Set thisNode = mAllNodes(pNodeKey)
        If mAllNodes.SelectedNode.IsDescendentOf(pNodeKey) Then mAllNodes.SelectNode pNodeKey
    End If
    
    If thisNode.Expanded And thisNode.Children.Count > 0 Then
        Debug.Print "COLLAPSING "; thisNode.Caption
    
        If thisNode.Key <> MASTER_ROOT Then thisNode.Expanded = False
    
        CascadeRemove pNodeKey, thisNode.Children, pCascade, lNodesConcealedCount
        
        If mRecalcWidestNode Then CalcMaxNodeWidth
        
        RaiseEvent NodesRemoved
        
        Debug.Print "COLLAPSE COMPLETE: " & lNodesConcealedCount & " nodes concealed"
    End If

End Sub
Friend Sub Rebuild()
'Performs a complete rebuild of the mExposedNodes collection.
'Used when it is impractical to employ animation or no more efficient than working out what's been added or removed from it
   Set mExposedNodes = Nothing
   Set mExposedNodes = New Collection
   'start recursive call to identify all the nodes
   mRecalcWidestNode = True
   RebuildTreeWalk mAllNodes.RootNode.Children
   RaiseEvent CompletelyRebuilt
End Sub
Private Function RebuildTreeWalk(pTheseNodes As Collection)
Dim thisNode As cVhNode
    For Each thisNode In pTheseNodes
        Add thisNode
        If thisNode.Expanded Then RebuildTreeWalk thisNode.Children
    Next thisNode
End Function
Private Function CascadeAdd(pExpandingNodeKey As String, _
                        pInsertAfterKey As String, _
                        pTheseNodes As Collection, _
                        pChildrenAreExposed As Boolean, _
                        pCascadeExpand As Boolean, _
                        pNodesRevealedCount As Long, _
                        pVisibleDescendentCount As Long)
Dim thisNode As cVhNode, bNodeIsExposed As Boolean, bNodeIsExpanded As Boolean
    
    For Each thisNode In pTheseNodes
         thisNode.SortChildren

        bNodeIsExposed = pChildrenAreExposed
        bNodeIsExpanded = thisNode.Expanded
        
        If bNodeIsExposed Then
            If Not bNodeIsExpanded Then bNodeIsExposed = False
        Else
            Add thisNode, pInsertAfterKey
            pNodesRevealedCount = pNodesRevealedCount + 1
            Animate pNodesRevealedCount, pExpandingNodeKey, pVisibleDescendentCount
        End If
        
        pInsertAfterKey = thisNode.Key
        
        pVisibleDescendentCount = pVisibleDescendentCount + 1
        If pCascadeExpand Then thisNode.Expanded = True
        
        If thisNode.Expanded Then
            CascadeAdd pExpandingNodeKey, pInsertAfterKey, thisNode.Children, bNodeIsExposed, pCascadeExpand, pNodesRevealedCount, pVisibleDescendentCount
        End If
        
    Next thisNode
    
End Function
Private Function CascadeRemove(pCollapsingNodeKey As String, _
                                        pTheseNodes As Collection, _
                                        pCascadeCollapse As Boolean, _
                                        pNodesConcealedCount As Long)
Dim thisNode As cVhNode
    
    For Each thisNode In pTheseNodes
        
        If thisNode.Tier > 1 Then
            Remove thisNode
            pNodesConcealedCount = pNodesConcealedCount + 1
        End If
        
        Animate pNodesConcealedCount
        
        If thisNode.Expanded Then CascadeRemove pCollapsingNodeKey, thisNode.Children, pCascadeCollapse, pNodesConcealedCount
    
        If pCascadeCollapse Then thisNode.Expanded = False
    
    Next thisNode
    
End Function
Private Sub Animate(pIteration As Long, Optional pExpandingNodeKey As String = vbNullString, Optional pVisibleDescendentCount As Long = 0)
Dim bExpanding As Boolean
Static NextNotification As Double

    If Not mNotifyFrequency > 0 Then Exit Sub 'animation is OFF; quit...
    If pIteration = 0 Then Exit Sub 'nothing to do (calling sub didn't do anything)
    
    If pIteration = 1 Then NextNotification = 1
    
    If NextNotification > 1024 Then Exit Sub 'we don't want it to get too big!
    
    If pIteration Mod (NextNotification \ 1) = 0 Then 'time to notify the Viewport

        NextNotification = NextNotification * mNotifyFrequency 'we'll increase the value of l now
        
        bExpanding = Len(pExpandingNodeKey) > 0
        
        If bExpanding Then
            RaiseEvent NodesAdded(pExpandingNodeKey, pVisibleDescendentCount)
        Else
            RaiseEvent NodesRemoved
        End If
        Debug.Print "...Notifying change to Viewport at iteration: "; pIteration
    End If
End Sub

Private Sub Class_Initialize()
    Set mAllNodes = New cVhNodes
    Set mExposedNodes = New Collection
    mRecalcWidestNode = True
    mNodeIndentation = 16
End Sub

Private Sub Class_Terminate()
   On Error Resume Next
   Set mAllNodes = Nothing
   Set mExposedNodes = Nothing
End Sub
Friend Property Get AllNodes() As cVhNodes
    Set AllNodes = mAllNodes
End Property
Friend Property Let NotifyIncrementally(pFrequency As Double)
    mNotifyFrequency = pFrequency
End Property
Friend Property Get Count() As Long
    Count = mExposedNodes.Count
End Property
Friend Property Get Items() As Collection
    Set Items = mExposedNodes
End Property
Friend Property Let NodeIndentation(pIndentation As Double)
    mNodeIndentation = pIndentation
End Property
Friend Property Get WidestNodeWidth() As Long
    WidestNodeWidth = mWidestNodeWidth
End Property
Private Sub Add(pNode As cVhNode, Optional pAfter As String = vbNullString)
Dim NodeWidth As Double

    If Len(pAfter) > 0 Then
        mExposedNodes.Add pNode, pNode.Key, , pAfter
    Else
        mExposedNodes.Add pNode, pNode.Key
    End If
    
    NodeWidth = (((pNode.Tier - 1) * mNodeIndentation) + pNode.CaptionWidth)
    If NodeWidth > mWidestNodeWidth Then
        mWidestNodeWidth = NodeWidth
        mWidestNodeKey = pNode.Key
    End If
    
End Sub
Private Sub Remove(pNode As cVhNode)
    If mWidestNodeKey = pNode.Key Then
        mRecalcWidestNode = True
    End If
    mExposedNodes.Remove pNode.Key
End Sub
Friend Sub CalcMaxNodeWidth()
Dim thisNode As cVhNode, thisNodeWidth As Long

    Debug.Print "Calc MAX Node Width"

    'painful this, but need to measure 'virtual' width of control
    mWidestNodeWidth = 0
    mRecalcWidestNode = False
    For Each thisNode In mExposedNodes
        thisNodeWidth = (((thisNode.Tier - 1) * mNodeIndentation) + thisNode.CaptionWidth)
        If thisNodeWidth > mWidestNodeWidth Then
            mWidestNodeWidth = thisNodeWidth
            mWidestNodeKey = thisNode.Key
        End If
    Next thisNode

End Sub

