VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "cTreeViewport"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

Private WithEvents mExposedNodes As cExposedNodes
Attribute mExposedNodes.VB_VarHelpID = -1

Private mExactCapacity              As Double 'e.g. 17.25 nodes visible
Private mCapacity                   As Long  'e.g. 17 nodes
Private mPosition                   As Long 'the index within mExposedNodes of the first(top) node in the viewport
Private mUnderlyingItemCount        As Long
Private mAnimated                   As Boolean

Event ViewportChanged()
Event SelectionChanged(newPosition As Long)

Private Sub mExposedNodes_NodesAdded(pParentNodeKey As String, pExposedDescendentCount As Long)
'This event is fired by clsExposedNodes whenever a node is added to its mExposednodes collection
'i.e.   - when a node is expanded
'       - when a new node is created
Dim i As Long, ViewportIndex As Long, ShiftAmount As Long, OldAnimate As Boolean

    For i = mPosition To mPosition + VisibleItemCount - 1
        If UnderlyingItems(i).Key = pParentNodeKey Then
            ViewportIndex = i - mPosition + 1
            Exit For
        End If
    Next i
    
    If ViewportIndex > 0 Then
        'expanded item is visible in the viewport
        If ViewportIndex + pExposedDescendentCount < VisibleItemCount Then
            'descendents don't need to be scrolled into view - just update the control
            Refresh
        Else
            If ViewportIndex = 1 Then
               ShiftAmount = 0
            Else
               ShiftAmount = ViewportIndex + pExposedDescendentCount - Int(mExactCapacity)
            End If
            
            If ShiftAmount > 0 Then
                If ViewportIndex - ShiftAmount <= 0 Then ShiftAmount = ViewportIndex - 1
                Refresh
                'Turn off animation temporarily as 'Shift' has its own animation and we don't want to compound it
                OldAnimate = mAnimated
                mAnimated = False
                Shift ShiftAmount
                mAnimated = OldAnimate
            Else
                Refresh
            End If
            
        End If
    Else
        'need something here to handle when not in the viewport?
        Refresh
    End If
    
End Sub
Private Sub mExposedNodes_NodesRemoved()
    Refresh
End Sub
Friend Property Get AllNodes() As cVhNodes
    Set AllNodes = mExposedNodes.AllNodes
End Property
Friend Sub Refresh()
   mUnderlyingItemCount = mExposedNodes.Count
   
   If mUnderlyingItemCount = 0 Then mPosition = 1
   
   If mPosition > MaxPosition Then mPosition = MaxPosition
   
   'trigger a refresh
   RaiseEvent ViewportChanged
End Sub
Friend Property Let Capacity(pNumNodes As Double)
Dim NewCapacity As Long, CapacityChanged As Boolean

    mExactCapacity = pNumNodes

    If Int(mExactCapacity) = mExactCapacity Then
        NewCapacity = mExactCapacity
    Else
        NewCapacity = Int(mExactCapacity + 1)
    End If

    CapacityChanged = Not (mCapacity = NewCapacity)
    mCapacity = NewCapacity
    
    If CapacityChanged Then Refresh
    
End Property
Friend Property Get Capacity() As Double
   Capacity = mCapacity
End Property
Friend Property Get ExactCapacity() As Double
    ExactCapacity = mExactCapacity
End Property
Friend Property Let Position(pExposedIndex As Long)
    If pExposedIndex > MaxPosition Then
        pExposedIndex = MaxPosition
    ElseIf pExposedIndex < 1 Then
        pExposedIndex = 1
    End If
    
    If mPosition = pExposedIndex Then Exit Property
    
    mPosition = pExposedIndex
    
    RaiseEvent ViewportChanged

    Debug.Print "New Viewport Position = "; mPosition
End Property
Friend Sub Shift(pShiftAmount As Long)
Dim i As Long, j As Long, MovesCompleted As Long, MovesRemaining As Long

    If pShiftAmount = 0 Then Exit Sub 'nothing to do...
    
    If mAnimated Then
        j = 1
        For i = 1 To Abs(pShiftAmount)
            j = j * 2
            If MovesCompleted + j < Abs(pShiftAmount) And j < 1024 Then
                MovesCompleted = MovesCompleted + j
                Position = mPosition + (j * (pShiftAmount / Abs(pShiftAmount)))
            Else
                Exit For
            End If
        Next i
        MovesRemaining = Abs(pShiftAmount) - MovesCompleted
        If MovesRemaining > 0 Then
            Position = mPosition + (MovesRemaining * (pShiftAmount / Abs(pShiftAmount)))
        End If
    Else
        Position = mPosition + pShiftAmount
    End If
    'Debug.Print "Viewport shifted " & IIf(pShiftAmount < 0, "UP ", "DOWN ") & pShiftAmount
End Sub
Friend Property Get Position() As Long
    If mPosition > MaxPosition Then mPosition = MaxPosition
    Position = mPosition
End Property
Friend Property Get MaxPosition() As Long
    MaxPosition = 1
    If mUnderlyingItemCount - mCapacity >= 0 Then
        MaxPosition = 1 + mUnderlyingItemCount - (Int(mExactCapacity))
    End If
End Property
Friend Property Get VisibleItemCount() As Long
'the number of nodes that appear in the viewport (ultimately, client area of the control)
    If mUnderlyingItemCount = 0 Then
        VisibleItemCount = 0
    Else
        VisibleItemCount = mUnderlyingItemCount - mPosition + 1
    End If
    If VisibleItemCount > mCapacity Then
        VisibleItemCount = mCapacity
    End If
End Property

Private Sub Class_Initialize()
   Set mExposedNodes = New cExposedNodes
   AnimationSpeed = 1.5
   mPosition = 1
   mExactCapacity = 5
End Sub
Private Sub Class_Terminate()
   On Error Resume Next
   Set mExposedNodes = Nothing
End Sub
Friend Property Get KeyFromViewportIndex(pIndex As Long) As String
   KeyFromViewportIndex = UnderlyingItems(mPosition + pIndex - 1).Key
End Property
Friend Sub SelectNodeByViewportPosition(pViewportIndex As Long)
Dim thisNode As cVhNode
   If pViewportIndex > VisibleItemCount Then pViewportIndex = VisibleItemCount
   Set thisNode = UnderlyingItems(mPosition + pViewportIndex - 1)
   mExposedNodes.AllNodes.SelectNode thisNode.Key
   RaiseEvent SelectionChanged(pViewportIndex)
End Sub
Friend Sub EnsureVisible(pNodeKey As String)
Dim UnderlyingItemsIndex As Long, FullyVisibleItemCount As Long, thisNode As cVhNode
    
    For Each thisNode In UnderlyingItems
        UnderlyingItemsIndex = UnderlyingItemsIndex + 1
        If thisNode.Key = pNodeKey Then Exit For
    Next thisNode
    
    FullyVisibleItemCount = VisibleItemCount - IIf(mCapacity > mExactCapacity, 1, 0)
    If UnderlyingItemsIndex < mPosition Then
        Shift UnderlyingItemsIndex - mPosition
    ElseIf UnderlyingItemsIndex > (mPosition + FullyVisibleItemCount - 1) Then
        Shift UnderlyingItemsIndex - (mPosition + FullyVisibleItemCount - 1)
    Else
        RaiseEvent ViewportChanged
    End If
    
End Sub
Friend Property Let AnimationSpeed(pSpeed As Double)
    mAnimated = (pSpeed >= 1)
    mExposedNodes.NotifyIncrementally = pSpeed
End Property
Friend Property Get UnderlyingItems() As Collection
    Set UnderlyingItems = mExposedNodes.Items
End Property
Friend Property Get Item(pViewportPosition As Long) As cVhNode
    Set Item = mExposedNodes.Items(pViewportPosition + mPosition - 1)
End Property
Friend Property Get ExposedNodes() As cExposedNodes
    Set ExposedNodes = mExposedNodes
End Property

