VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "cVhNode"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

Private mCaption            As String
Private mCaptionColour      As Long
Private mType               As cNodeType
Private mCaptionWidth       As Double
Private mRoot               As cVhNode
Private mParent             As cVhNode
Private mKey                As String
Private mExpanded           As Boolean
Private mChildNodes         As Collection
Private mTier               As Integer

Event NodeExpanded(Key As String, Cascade As Boolean)
Event NodeCollapsed(Key As String, Cascade As Boolean)
Event NodeSorted()
Event ParentChanged(Key As String, NewParentKey As String)
Event CaptionChanged(NodeKey As String)
Private Function Initialised() As Boolean
   Initialised = Len(mCaption) > 0
End Function
Friend Property Set Root(pRoot As cVhNode)
   Set mRoot = pRoot
End Property
Friend Property Set Parent(pParent As cVhNode)
   Set mParent = pParent
End Property
Public Function SetParent(pNewParent As cVhNode) As Boolean
'Called when a new parent is assigned to a node
   Select Case True
      Case mParent.Key = pNewParent.Key
         Err.Raise 60000 Or vbObjectError, "cVhNode:SetParent", "Move Aborted - Node is already a child of this parent"
      Case mKey = pNewParent.Key
         Err.Raise 60001 Or vbObjectError, "cVhNode:SetParent", "Move Aborted - A node cannot be a child of itself"
      Case pNewParent.IsDescendentOf(mKey)
         Err.Raise 60002 Or vbObjectError, "cVhNode:SetParent", "Move Aborted - A node cannot be the child of a descendent"
      Case pNewParent.NodeType.IsValidChildType(mType.Name) = False
         Err.Raise 60003 Or vbObjectError, "cVhNode:SetParent", "Move Aborted - Nodes of type '" & pNewParent.NodeType & "' cannot contain nodes of type '" & mType & "'"
      Case Else
         If IsUniqueChild(pNewParent, mCaption, mType.Name) Then
            mParent.Children.Remove mKey
            pNewParent.Children.Add Me, mKey
            pNewParent.Expanded = True
            
            CascadeNewTier mChildNodes, pNewParent.Tier + 1 - mTier
            
            Set mParent = pNewParent
            mTier = mParent.Tier + 1
            SortParent
            
            mRoot.NotifyParentChange mKey, pNewParent.Key
            
            SetParent = True
         Else
            Err.Raise 60003 Or vbObjectError, "cVhNode:SetParent", "Move Aborted - '" & pNewParent.Caption & "' already contains a " & mType.Name & " called '" & mCaption & "'"
         End If
   End Select
End Function
Private Function CascadeNewTier(pChildNodes As Collection, pTierAdjustment As Integer)
Dim thisNode As cVhNode
   For Each thisNode In pChildNodes
      thisNode.Tier = thisNode.Tier + pTierAdjustment
      CascadeNewTier thisNode.Children, pTierAdjustment
   Next thisNode
End Function
Public Property Get Parent() As cVhNode
    Set Parent = mParent
End Property
Public Sub Expand(pCascade As Boolean)
   SortChildren
   mRoot.NotifyExpansion mKey, pCascade
End Sub
Public Sub Collapse(pCascade As Boolean)
   mRoot.NotifyCollapse mKey, pCascade
End Sub
Public Function FullPath() As String
Dim thisNode As cVhNode
   Set thisNode = Me
   Do Until thisNode.Tier = 0
      FullPath = thisNode.Caption & "\" & FullPath
      Set thisNode = thisNode.Parent
   Loop
End Function
Public Function NextSibling() As cVhNode
Dim thisNode As cVhNode, bFoundMe As Boolean
   For Each thisNode In Me.Parent.Children
      If bFoundMe Then
         Set NextSibling = thisNode
         Exit For
      End If
      If thisNode.Key = Me.Key Then bFoundMe = True
   Next thisNode
End Function
Public Function PreviousSibling() As cVhNode
Dim thisNode As cVhNode
   For Each thisNode In Me.Parent.Children
      If thisNode.Key = Me.Key Then Exit For
      Set PreviousSibling = thisNode
   Next thisNode
End Function
Public Property Set NodeType(pType As cNodeType)
   If Not (pType.Name = mType.Name) Then
      Set mType = pType
      If Initialised Then mType.MasterType.NotifyNodeTypeChange True, mKey, mParent.NodeType.SortMethod > eNoSort
   End If
End Property
Public Property Get NodeType() As cNodeType
    Set NodeType = mType
End Property
Friend Property Get DragDropBehaviour() As enumDragScope
    DragDropBehaviour = mType.DragDropBehaviour
End Property
Public Property Get CaptionColour() As Long
   If mCaptionColour = -1 Then
      CaptionColour = mType.CaptionColour
   Else
      CaptionColour = mCaptionColour
   End If
End Property
Public Property Let CaptionColour(pColour As Long)
   mCaptionColour = pColour 'if we set a colour (thereby over-riding the default value of -1), we can specify a colour for this node independent of its node type
End Property
Public Property Get Tier() As Integer
   Tier = mTier
End Property
Friend Property Let Tier(pTier As Integer)
   mTier = pTier
End Property
Public Property Get Expanded() As Boolean
   Expanded = mExpanded And (mChildNodes.Count > 0)
End Property
Friend Property Let Expanded(pBool As Boolean)
   mExpanded = pBool 'And (mChildNodes.Count > 0)
End Property
Friend Property Get CurrentImageKey() As String
   If mExpanded And mChildNodes.Count > 0 Then
      CurrentImageKey = mType.ExpandedImageKey
   Else
      CurrentImageKey = mType.ImageKey
   End If
End Property
Public Property Let Key(pKey As String)
   mKey = pKey
End Property
Public Property Get Key() As String
   Key = mKey
End Property
Public Property Let Caption(pCaption As String)
   
   If mCaption = pCaption Then Exit Property
   
   If Len(pCaption) = 0 Then 'get a unique name
      pCaption = mParent.GetUniqueChildName("New " & mType.Name, mType.Name)
   Else
      If Not IsUniqueChild(mParent, pCaption, mType.Name) Then
          Err.Raise 60004 Or vbObjectError, "cVhNode:Set Caption", "Denied - '" & mParent.Caption & "' already contains a " & mType.Name & " called '" & pCaption & "'"
      End If
   End If
   mCaption = pCaption
   If Not mKey = MASTER_ROOT Then mRoot.NotifyCaptionChanged mKey

End Property
Friend Property Get CaptionWidth() As Double
   CaptionWidth = mCaptionWidth
End Property
Friend Property Let CaptionWidth(pCaptionWidth As Double)
   mCaptionWidth = pCaptionWidth
End Property
Public Property Get Caption() As String
   Caption = mCaption
End Property
Public Function AddChild(pChild As cVhNode)
   mChildNodes.Add pChild, pChild.Key
End Function
Public Property Get Children() As Collection
   Set Children = mChildNodes
End Property
Friend Property Get IsValidInternalDragSource() As Boolean
   IsValidInternalDragSource = mType.DragDropBehaviour And eInternalSource
End Property
Friend Property Get IsValidExternalDragSource() As Boolean
   IsValidExternalDragSource = mType.DragDropBehaviour And eExternalSource
End Property
Friend Property Get IsValidInternalDropTarget() As Boolean
   IsValidInternalDropTarget = mType.DragDropBehaviour And eInternalTarget
End Property
Friend Property Get IsValidExternalDropTarget() As Boolean
   IsValidExternalDropTarget = mType.DragDropBehaviour And eExternalTarget
End Property
Public Sub EnsureVisible()
Dim thisNode As cVhNode
   Set thisNode = OldestExposedAncestor
   Do Until thisNode.Key = Me.Key
      thisNode.Expand False
      Set thisNode = OldestExposedAncestor
   Loop
End Sub
Private Function OldestExposedAncestor() As cVhNode
Dim thisNode As cVhNode
   Set thisNode = Me
   Do Until thisNode.IsExposed '.Parent.Expanded
       Set thisNode = thisNode.Parent
   Loop
   Set OldestExposedAncestor = thisNode
End Function
Public Function IsExposed() As Boolean
Dim thisNode As cVhNode
   Set thisNode = Me
   Do Until thisNode.Tier = 0
      If Not thisNode.Parent.Expanded Then Exit Do
      Set thisNode = thisNode.Parent
   Loop
   IsExposed = thisNode.Tier = 0
End Function
Public Function IsDescendentOf(pAncestorKey As String) As Boolean
Dim ParentNode As cVhNode
   If mKey = MASTER_ROOT Then Exit Function
   Set ParentNode = mParent
   Do Until ParentNode.Tier < 1
      If ParentNode.Key = pAncestorKey Then
         IsDescendentOf = True
         Exit Do
      End If
      Set ParentNode = ParentNode.Parent
   Loop
End Function
Public Function IsUniqueChild(pParentNode As cVhNode, pTestCaption As String, pNodeType As String)
Dim thisNode As cVhNode
   IsUniqueChild = True
   If mKey = "-|*MASTER*|-" Then Exit Function
   For Each thisNode In pParentNode.Children
      If thisNode.Key <> mKey Then
         If thisNode.Caption = pTestCaption Then
            If thisNode.NodeType.Name = pNodeType Then
               IsUniqueChild = False
               Exit For
            End If
         End If
      End If
   Next thisNode
End Function
Friend Function GetUniqueChildName(pPrefix As String, pNodeType As String) As String
Dim i As Long, Found As Boolean, Suffix As String
   Do
      If i > 0 Then Suffix = " (" & i & ")"
      Found = IsUniqueChild(Me, pPrefix & Suffix, pNodeType)
      i = i + 1
      If i = 1 Then i = 2
   Loop Until Found
   GetUniqueChildName = pPrefix & Suffix
End Function
Private Sub SortParent()
   If Not mParent Is Nothing Then mParent.Sort
End Sub
Public Sub Sort()
   If Me.Children.Count > 0 Then
      If mType.SortMethod > eNoSort Then
         SortChildren
         If mKey <> "-|*MASTER*|-" Then
            mRoot.NotifySort
         Else
            NotifySort
         End If
      End If
   End If
End Sub
Friend Sub SortChildren() 'ADVICE NEEDED: would welcome a faster way of doing this!
Dim oOldNode As cVhNode, oNewNode As cVhNode
Dim colNewChildren As Collection
Dim lngIdx As Long
   
   Set colNewChildren = New Collection
   
   For Each oOldNode In mChildNodes
      If colNewChildren.Count = 0 Then
         colNewChildren.Add oOldNode, oOldNode.Key
      Else
         For lngIdx = 1 To colNewChildren.Count
            Set oNewNode = colNewChildren(lngIdx)
            If ComesBefore(oOldNode, oNewNode) Then
               colNewChildren.Add oOldNode, oOldNode.Key, lngIdx
               Exit For
            End If
         Next
         If lngIdx > colNewChildren.Count Then
            colNewChildren.Add oOldNode, oOldNode.Key
         End If
      End If
   Next
   
   Set mChildNodes = colNewChildren
   Set colNewChildren = Nothing
   
   'If mChildNodes.Count > 0 Then Debug.Print "Sorted " & mCaption
   
End Sub
Private Function ComesBefore(TestNode As cVhNode, CompareToNode As cVhNode) As Boolean
   Select Case True
      Case mType.SortMethod = eCaption
         If StrComp(TestNode.Caption, CompareToNode.Caption, vbTextCompare) < 1 Then
            ComesBefore = True
         End If
      Case mType.SortMethod = ePriority
         If TestNode.NodeType.SortPriority < CompareToNode.NodeType.SortPriority Then
            ComesBefore = True
         ElseIf TestNode.NodeType.SortPriority = CompareToNode.NodeType.SortPriority Then
            If StrComp(TestNode.Caption, CompareToNode.Caption, vbTextCompare) < 1 Then
               ComesBefore = True
            End If
         End If
   End Select
End Function
'only the Master Node, which is referenced by all other nodes, will ever raise the 'Notify' events
Friend Sub NotifyExpansion(pKey As String, pCascade As Boolean)
   RaiseEvent NodeExpanded(pKey, pCascade)
End Sub
Friend Sub NotifyCollapse(pKey As String, pCascade As Boolean)
   RaiseEvent NodeCollapsed(pKey, pCascade)
End Sub
Friend Sub NotifyParentChange(pKey As String, pNewParentKey As String)
   RaiseEvent ParentChanged(pKey, pNewParentKey)
End Sub
Friend Sub NotifySort()
   RaiseEvent NodeSorted
End Sub
Friend Sub NotifyCaptionChanged(pKey As String)
   RaiseEvent CaptionChanged(pKey)
End Sub
Private Sub Class_Initialize()
   Set mChildNodes = New Collection
   Set mType = New cNodeType
   mCaptionColour = -1 'when -1, the caption colour for the node's type will be used
End Sub
Private Sub Class_Terminate()
   On Error Resume Next
   Set mChildNodes = Nothing
   Set mType = Nothing
   Set mRoot = Nothing
   Set mParent = Nothing
End Sub
