VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "cVhNodes"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

Private mNodes                      As Collection 'all the nodes within the tree
Private WithEvents mNodeTypes       As cNodeTypes
Attribute mNodeTypes.VB_VarHelpID = -1
'the 'master' node for the tree, whose key is MASTER_ROOT; implicitly owns all items in the mNodes collection but is never displayed
'Also used as a conduit to raise events through since all nodes hold a reference to it
Private WithEvents mRootNode        As cVhNode
Attribute mRootNode.VB_VarHelpID = -1
Private mSelectedNodeKey            As String
Private mSuspended                  As Boolean 'used to cancel event raising during intensive internal tasks (like sorting)

'Events raised to clsExposedNodes so that it can refresh itself
Event NodeExpanded(NodeKey As String, CascadeExpansion As Boolean)
Event NodeCollapsed(NodeKey As String, CascadeCollapse As Boolean)
Event NodeAdded(NewNode As cVhNode)
Event NodeRemoved(RemovedNode As cVhNode)
Event NodeMoved()
Event NodesSorted()
Event NodesCleared()
Event NodeTypeChanged(RecalcNeeded As Boolean, NodeKey As String)
Event NodeCaptionChanged(NodeKey As String)
Private Sub mNodeTypes_NodeTypeChanged(RecalcNeeded As Boolean, NodeKey As String, SortNeeded As Boolean)
   If Not mSuspended Then
      If SortNeeded Then
         If Len(NodeKey) > 0 Then
            mNodes(NodeKey).Parent.Sort
         Else
            SortAll
            RaiseEvent NodesSorted
         End If
      Else
         RaiseEvent NodeTypeChanged(RecalcNeeded, NodeKey)
      End If
   End If
End Sub
Private Sub mRootNode_CaptionChanged(NodeKey As String)
   If Not mSuspended Then
      mNodes(NodeKey).Parent.Sort
      RaiseEvent NodeCaptionChanged(NodeKey)
   End If
End Sub
Private Sub mRootNode_NodeCollapsed(Key As String, Cascade As Boolean)
    RaiseEvent NodeCollapsed(Key, Cascade)
End Sub
Private Sub mRootNode_NodeExpanded(Key As String, Cascade As Boolean)
    RaiseEvent NodeExpanded(Key, Cascade)
End Sub
Private Sub mRootNode_NodeSorted()
   If Not mSuspended Then RaiseEvent NodesSorted
End Sub
Private Sub mRootNode_ParentChanged(Key As String, NewParentKey As String)
    RaiseEvent NodeMoved
End Sub
'NODE SELECTION
Friend Property Get SelectedNode() As cVhNode
On Error GoTo ErrorHandler
   Set SelectedNode = mNodes(mSelectedNodeKey)
   Exit Property
ErrorHandler:
   If mNodes.Count > 0 Then
      Set SelectedNode = mNodes(1)
   Else
      Set SelectedNode = Nothing
   End If
End Property
Public Sub SelectNode(pIndex As Variant)

   If mNodes.Count = 0 Then Exit Sub
   
   If TypeName(pIndex) = "Long" Then
      If pIndex < 1 Then pIndex = 1
   End If
   
   If pIndex = MASTER_ROOT Then
      mSelectedNodeKey = MASTER_ROOT
   Else
      mSelectedNodeKey = mNodes(pIndex).Key
   End If

   'Debug.Print SelectedNode.Caption; " was selected: Key is "; SelectedNode.Key
End Sub

'NODE ADDITION
Public Sub Add(pParentKey As String, pKey As String, pCaption As String, pType As String, Optional pCaptionColour As Long = -1)
Dim NewNode As cVhNode, ParentNode As cVhNode, CurrentSortMethod As enumNodeSortMethod
    
On Error GoTo ErrorHandler

   If Exists(pKey) Then
      Err.Raise 457, "cVhNode:Add", "Node Key Already Exists"
   End If
   
   If Len(pParentKey) = 0 Or (pParentKey = MASTER_ROOT) Then
      Set ParentNode = mRootNode
   Else
      Set ParentNode = mNodes(pParentKey)
   End If
    
   If Not ParentNode.NodeType.IsValidChildType(pType) Then
      Err.Raise 60010 Or vbObjectError, "cVhNode:Add", "The parent node type does not allow this node type as a child"
   End If
    
   If mSuspended Then
      'if the control is suspended we don't want new nodes to trigger re-sorts
      'especially if we're bulk-populating a tree (e.g. app startup)
      'the easiest way to turn this off is to change the sort method for the Node Type
      CurrentSortMethod = mNodeTypes(pType).SortMethod
      mNodeTypes(pType).SortMethod = eNoSort
   End If
    
   'Create a new node...
   Set NewNode = New cVhNode
   With NewNode
      .Key = pKey
      .Tier = ParentNode.Tier + 1
      Set .NodeType = mNodeTypes(pType)
      Set .Root = mRootNode
      Set .Parent = ParentNode
      
      '...and add it to the main nodes collection
      mNodes.Add NewNode, pKey
      'add it, also, to it's parent's collection of nodes, so that it's with its siblings. Aaaaah
      ParentNode.AddChild NewNode
      
      .CaptionColour = pCaptionColour
      
      'need to do this last after it's been added to is parent as caption setting triggers a sort
      .Caption = pCaption
   End With
   
   
   If mSuspended Then
      mNodeTypes(pType).SortMethod = CurrentSortMethod 'restore the Type's sort method...
   Else
      RaiseEvent NodeAdded(NewNode)
   End If

   Exit Sub
    
ErrorHandler:

    If Not Err.Number = 457 Then
'        mNodes.Remove pKey
'        ParentNode.Children.Remove pKey
    End If
    Err.Raise Err.Number, Err.Source, Err.Description
End Sub

'NODE REMOVAL
Public Sub Remove(pIndex As Variant)
Dim thisNode As cVhNode, NewSelectedNode As cVhNode

    Set thisNode = mNodes(pIndex)
    
    'work out what node should be selected after this node is removed
    If thisNode.Parent.Children.Count > 1 Then
        Set NewSelectedNode = thisNode.NextSibling
        If NewSelectedNode Is Nothing Then
            Set NewSelectedNode = thisNode.PreviousSibling
        End If
    Else
        Set NewSelectedNode = thisNode.Parent
    End If
    
    If Not mSuspended Then RaiseEvent NodeRemoved(thisNode)
    
    'Now we can remove it from its siblings via its parent and also from the main mNodes collection
    thisNode.Parent.Children.Remove thisNode.Key
    mNodes.Remove thisNode.Key
    CascadeRemove thisNode.Children
    
    'Set the new SelectedNode we worked out earlier
    SelectNode NewSelectedNode.Key
    
End Sub
Private Function CascadeRemove(pChildNodes As Collection)
Dim thisNode As cVhNode
   For Each thisNode In pChildNodes
      thisNode.Parent.Children.Remove thisNode.Key
      mNodes.Remove thisNode.Key
      CascadeRemove thisNode.Children
   Next thisNode
End Function

'NODE EXPANSION AND COLLAPSING
Public Sub ExpandAll()
   RaiseEvent NodeExpanded(MASTER_ROOT, True)
End Sub
Public Sub CollapseAll()
   RaiseEvent NodeCollapsed(MASTER_ROOT, True)
End Sub
Friend Sub SortAll()
Dim thisNode As cVhNode
   mSuspended = True
   mRootNode.Sort
   New_c.Timing True
   For Each thisNode In mNodes
      If thisNode.Expanded Then thisNode.SortChildren
   Next thisNode
   Debug.Print "Sort took " & New_c.Timing
   mSuspended = False
   RaiseEvent NodesSorted
End Sub

'UTILITY FUNCTIONS/SUBS
Public Property Get Item(pIndex As Variant) As cVhNode ' this is the default property
Attribute Item.VB_UserMemId = 0
On Error GoTo ErrorHandler
   Set Item = mNodes.Item(pIndex)
   Exit Property
ErrorHandler:
Err.Raise 35601, "cVhNode", "Element Not Found"
End Property
Public Function Exists(pIndex As Variant) As Boolean
Dim thisNode As cVhNode
   On Error Resume Next
   Set thisNode = mNodes.Item(pIndex)
   Exists = Err.Number = 0
End Function
Public Sub Clear()
   CascadeRemove mRootNode.Children
   Set mNodes = New Collection
   mSelectedNodeKey = vbNullString
   RaiseEvent NodesCleared
End Sub
'PUBLIC PROPERTIES
Public Property Get Count() As Long
   Count = mNodes.Count
End Property
'FRIEND PROPERTIES AND FUNCTIONS
Friend Property Get NodeTypes() As cNodeTypes
   Set NodeTypes = mNodeTypes
End Property
Public Property Get RootNode() As cVhNode
   Set RootNode = mRootNode
End Property
Public Function NewEnum() As IUnknown
Attribute NewEnum.VB_UserMemId = -4
    Set NewEnum = mNodes.[_NewEnum]
End Function
Public Property Get Nodes() As Collection
   Set Nodes = mNodes
End Property
Friend Property Let Suspended(pValue As Boolean)
   If mSuspended = pValue Then Exit Property
   mSuspended = pValue
End Property
Friend Property Get Suspended() As Boolean
   Suspended = mSuspended
End Property
'CLASS CONSTRUCTOR/DESTRUCTOR
Private Sub Class_Initialize()
    Set mNodeTypes = New cNodeTypes
    Set mRootNode = New cVhNode
    With mRootNode
        .Key = MASTER_ROOT
        .Tier = 0
        .Expanded = True
        Set .NodeType = mNodeTypes(ROOT_TYPE)
        .Caption = "Root"
    End With
    Clear
End Sub
Private Sub Class_Terminate()
   On Error Resume Next
   Clear
   Set mRootNode = Nothing
   Set mNodes = Nothing
   Set mNodeTypes = Nothing
End Sub
 
